/*
 * misc.c - Miscelaneous helpers and wrappers.
 *
 * Grigoriy Sitkarev, <sitkarev@unixkomi.ru>
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <limits.h>
#include <errno.h>
#include <pwd.h>
#include <grp.h>
#include <ctype.h>

#include "xmalloc.h"
#include "log.h"

#define DEV_NULL	"/dev/null"

int str_isnumber(const char *str)
{
	if (!str || *str == 0)
		return 0;

	while (*str && isdigit(*str))
		str++;

	return (*str == 0);
}

void sanitize_stdio()
{
	int res;
	int i, devnull;

	devnull = open(DEV_NULL, O_RDWR);

	if (devnull == -1)
		error("open() '%s' failed: %s", DEV_NULL, strerror(errno));

	i = devnull;

	while (++i <= 2) {
		res = fcntl(i, F_GETFL);
		if (res >= 0)
			continue;
		res = dup2(devnull, i);
		if (res == -1)
			error("dup2() can't dup fd=%d: %s", i, strerror(errno));
	}

	if (devnull > 2)
		close(devnull);
}

void daemonize()
{
	int i, devnull;
	int res;
	pid_t pid;

	chdir("/");
	umask(0220);
	pid = fork();
	if (pid > 0)
		exit(0);
	if (pid < 0)
		error("fork() failed: %s", strerror(errno));
	setsid();
	devnull = open(DEV_NULL, O_RDWR | O_NOCTTY);
	if (devnull == -1)
		error("can't open() '%s': %s", DEV_NULL, strerror(errno));
	i = 0;
	do {
		res = close(i);
		if (res == -1)
			warn("close() failed: %s", strerror(errno));
		res = dup2(devnull, i);
		if (res == -1)
			warn("dup2() failed: %s", strerror(errno));
	} while (++i <= 2);
	close(devnull);
}

int getpw_by_name_or_uid(const char *str, struct passwd *pwd)
{
	struct passwd *ptr = NULL;

	if (!str)
		return -1;

	if (str_isnumber(str)) {
		do {
			errno = 0;
			ptr = getpwuid(atoi(str));
		} while (ptr == NULL && errno == EINTR);
	} else {
		do {
			errno = 0;
			ptr = getpwnam(str);
		} while (ptr == NULL && errno == EINTR);
	}
	if (ptr) {
		if (pwd != NULL)
			*pwd = *ptr;
		return 0;
	}
	return -1;
}

int getgr_by_name_or_gid(const char *str, struct group *grp)
{
	struct group *ptr = NULL;

	if (!str)
		return -1;

	if (str_isnumber(str)) {
		do {
			errno = 0;
			ptr = getgrgid(atoi(str));
		} while (ptr == NULL && errno == EINTR);
	} else {
		do {
			errno = 0;
			ptr = getgrnam(str);
		} while (ptr == NULL && errno == EINTR);
	}
	if (ptr) {
		if (grp != NULL)
			*grp = *ptr;
		return 0;
	}
	return -1;
}

/* helper to obtain canonical absolute path */
char *canonize_path(const char *str)
{
	size_t len;
	char *p, *q, *cwd, *res;
	char *path;

	if (!str)
		return NULL;

	if (*str == '/') {
		len = strlen(str);
		res = xmalloc0(len + 1);
		path = xstrdup(str);
		res[0] = '/';
		p = res;
	} else {
		cwd = xgetcwd();
		len = strlen(cwd) + strlen(str) + 2;
		res = xmalloc0(len);
		path = xmalloc0(len);
		snprintf(path, len, "%s/%s", cwd, str);
		p = res;
		xfree(cwd);
	}

	q = path;

	while (*q) {
		if (q[0] == '/') {
			if (q[1] && q[1] == '/') {
				q += 1;
				continue;
			}
			if (q[1] && q[1] == '.') {
				if (q[2] == 0 || q[2] == '/') {
					q += 2;
					continue;
				}
				if (q[2] && q[2] == '.') {
					if (q[3] == 0 || q[3] == '/') {
						if (p != res && *p == '/')
							*p-- = 0;
						while (p != res && *p != '/')
							*p-- = 0;
						q += 3;
						continue;
					}
				}
			}
		}
		*p++ = *q++;
	}

	xfree(path);

	return res;
}
